﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace HeuristicLab.Algorithms.Bandits {
  // also called softmax policy
  public class BoltzmannExplorationPolicy : BanditPolicy {
    private readonly Random random;
    private readonly double eps;
    private readonly int[] tries;
    private readonly double[] sumReward;
    private readonly double beta;

    public BoltzmannExplorationPolicy(Random random, int numActions, double beta)
      : base(numActions) {
      if (beta < 0) throw new ArgumentException();
      this.random = random;
      this.beta = beta;
      this.tries = new int[numActions];
      this.sumReward = new double[numActions];
    }

    public override int SelectAction() {
      Debug.Assert(Actions.Any());
      // select best
      var maxReward = double.NegativeInfinity;
      int bestAction = -1;
      if (Actions.Any(a => tries[a] == 0))
        return Actions.First(a => tries[a] == 0);

      var ts = Actions.Select(a => Math.Exp(beta * sumReward[a] / tries[a]));
      var r = random.NextDouble() * ts.Sum();

      var agg = 0.0;
      foreach (var p in Actions.Zip(ts, Tuple.Create)) {
        agg += p.Item2;
        if (agg >= r) return p.Item1;
      }
      throw new InvalidProgramException();
    }
    public override void UpdateReward(int action, double reward) {
      Debug.Assert(Actions.Contains(action));

      tries[action]++;
      sumReward[action] += reward;
    }

    public override void DisableAction(int action) {
      base.DisableAction(action);
      sumReward[action] = 0;
      tries[action] = -1;
    }

    public override void Reset() {
      base.Reset();
      Array.Clear(tries, 0, tries.Length);
      Array.Clear(sumReward, 0, sumReward.Length);
    }

    public override void PrintStats() {
      for (int i = 0; i < sumReward.Length; i++) {
        if (tries[i] >= 0) {
          Console.Write("{0,5:F2}", sumReward[i] / tries[i]);
        } else {
          Console.Write("{0,5}", "");
        }
      }
      Console.WriteLine();
    }

    public override string ToString() {
      return string.Format("BoltzmannExplorationPolicy({0:F2})", beta);
    }
  }
}
