﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace HeuristicLab.Algorithms.Bandits {
  /* see: Streeter and Smith: A simple distribution-free approach to the max k-armed bandit problem, Proceedings  of the 12th
International Conference, CP 2006, Nantes, France, September 25-29, 2006. pp 560-574 */

  public class ChernoffIntervalEstimationPolicy : BanditPolicy {
    private readonly int[] tries;
    private readonly double[] sumReward;
    private int totalTries = 0;
    private readonly double delta;

    public ChernoffIntervalEstimationPolicy(int numActions, double delta = 0.01)
      : base(numActions) {
      this.delta = delta;
      this.tries = new int[numActions];
      this.sumReward = new double[numActions];
    }

    public override int SelectAction() {
      int bestAction = -1;
      double bestQ = double.NegativeInfinity;
      double k = Actions.Count();
      Debug.Assert(k > 0);
      foreach (var a in Actions) {
        if (tries[a] == 0) return a;
        // page 5 of "A simple distribution-free appraoch to the max k-armed bandit problem"
        // var alpha = Math.Log(2 * totalTries * k / delta);
        double alpha = Math.Log(2) + Math.Log(totalTries) + Math.Log(k) - Math.Log(delta); // total tries is max tries in the original paper
        double mu = sumReward[a] / tries[a];
        var q = mu + (alpha + Math.Sqrt(2 * tries[a] * mu * alpha + alpha * alpha)) / tries[a];
        if (q > bestQ) {
          bestQ = q;
          bestAction = a;
        }
      }
      return bestAction;
    }
    public override void UpdateReward(int action, double reward) {
      Debug.Assert(Actions.Contains(action));
      totalTries++;
      tries[action]++;
      sumReward[action] += reward;
    }

    public override void DisableAction(int action) {
      base.DisableAction(action);
      totalTries -= tries[action];
      tries[action] = -1;
      sumReward[action] = 0;
    }

    public override void Reset() {
      base.Reset();
      totalTries = 0;
      Array.Clear(tries, 0, tries.Length);
      Array.Clear(sumReward, 0, sumReward.Length);
    }

    public override void PrintStats() {
      for (int i = 0; i < sumReward.Length; i++) {
        if (tries[i] >= 0) {
          Console.Write("{0,5:F2}", sumReward[i] / tries[i]);
        } else {
          Console.Write("{0,5}", "");
        }
      }
      Console.WriteLine();
    }
    public override string ToString() {
      return string.Format("ChernoffIntervalEstimationPolicy({0:F2})", delta);
    }
  }
}
